// alloc 1024 memory blocks and test for memory bit errors
//   free good memory blocks (for other applications)
//   block bad memory blocks by sleeping forever after freeing good ones
//     == blacklist bad memory block at runtime
// if some bits are bad, 0.1% of memory will be occupied and rest can be used
// alternativ: linux badram/badmem kernel-option
//  EDAC = Error Detection And Correction
//
// gcc -Wall -O2          -static -o mem_badram mem_badram.c # gcc-4.1--
// gcc -Wall -O2 -fopenmp -static -o mem_badram mem_badram.c # gcc-4.2++
//     # ToDo:  -funroll-loops -ftree-vectorize # comp2
//     # intel-cc: -openmp
//     # -fopenmp needs gcc version 4.2 or later
// usage: mem_badram [<megabyte=1000>] [-l<nloops=100>] [-v<verbose=0>]
// HELP: use cpu affinity to check parts of numa memory on sockets/CPUs
// HELP: disable ECC at bios may be helpful identifying bad ram faster
// HELP: BIOS Data Scrambling
//   see https://en.m.wikipedia.org/wiki/Memory_controller + [6]=NXP+[7]
//  The memory controller incorporates a DDR3 Data Scrambling feature to
//  minimize the impact of excessive di/dt on the platform DDR3 VRs due to
//  successive 1s and 0s on the data bus.  Past experience has demonstrated
//  that traffic on the data bus is not random and can have energy
//  concentrated at specific spectral harmonics creating high di/dt that is
//  generally limited by data patterns that excite resonance between the
//  package inductance and on-die capacitances.  As a result, the memory
//  controller uses a data scrambling feature to create pseudo-random
//  patterns on the DDR3 data bus to reduce the impact of any excessive
//  di/dt.
// GOMP_CPU_AFFINITY="0-15"  # works as expected (gomp-1,comp2)
// OMP_PROC_BIND=TRUE # thread bounding, thread affinity (dont work!?)
// OMP_WAIT_POLICY="PASSIVE" OMP_NUM_THREADS=2 ./mem_badram 7168 # 7GB test
// ./mem_badram 31000  # 31GB test, argument1 is physical memory in MB
//  OMP_NUM_THREADS=16 srun -p sw01_short -n 1 -t 00:10:00 ./mem_badram 31000 # /core
// for x in $(find /sys/devices/system/edac/ -name \*panic_on_ue);do
//   echo $x $(cat $x);
//   echo 0 > $x; done   # HELP: disable kernel panic on uncorrectable errors
//       # if you have UE-errors which panic your system, only some CE errors
//       #   may reported, so disabling panic_on_ue will detect more errors
// ToDo: a detect, but do not correct ECC mode would be a big advantage
//       Soft-ECC would be much more flexible than hardware ECC (or FPGA)
//  --task-mem=450 ** OpenMP fatal error: cannot allocate critical lock
//
// 2020-06-18 precise output GiB/s (old=MB/s but was MiB/s)
//            fix bad factor 2 speed on ecc half-block-test
//            problem: multicore gives overlapping addresses for blocks
//            t100 non-ecc + non-interleave tests with bad DIMM n017.D1
// 2018-04-04 improvements for slurm-cluster +ECC-RAM health scan
//            add ompi-support (use slurm or ompi to start on many nodes)
// 2015-10-28 change OMP-loop-vars to signed (avoid pathcc.omp warn+crash)
// 2015-09-04 better avoid in cache-loops, correct speed if malloc-fail
// 2015-08-31 +openmp, fix use of lrnd, +output speed
// 2013-03-22 improved version
// 2011-07-06 cosmetic changes for output (more info)
// 2013 better use memspeed --test=8 or 16  ???
//
// ToDo20: noECC get physical mem-address and DIMM-No from virtual address
// ToDo18: better binary half intervall search? but blk gt cache-size? 
//         or test single blocks until fail? 1024*1s?
//         type of error? on write, on read?
// ToDo18 w+r alles, better partly wait longer (skip r/writes) log2-ranges
//   to have row-hammer like pattern .....xxxxxx...xxx...xxx..xx..xx..xx.x.x.x.x
// ToDo: reread conf-file on signal? change used mem or idle/sleep-param
//       long-time tests, optional monitor CE/UE from EDAC (own statistics)
//       log-behav: rewrite+check half of mem double speed + reorder fast ranges
// ToDo: output option, -q quiet for multi-node-mode (getenv SLURM),
//       repeat for nh houres, memcpy or vector statt single copy?
//  OMP_SCHEDULE=dynamic,1 # default
//  OMP_SCHEDULE=static,1  # test
// ToDo: slurm shared? sample job-script
// ToDo: disable ECC in BIOS for tests? Asus=yes(ibio)
//       lsmod | grep edac # sb_edac edac_core
//       cat /sys/devices/system/edac/mc/mc[01]/[uc]e_*count|xargs
//       https://www.kernel.org/doc/Documentation/edac.txt
//   dmidecode 64bit = internal ECC only (if so)
//             72bit = external ECC on Board or CPU(+memctl) or FakeECC 
//    k2.6.16 /sys/devices/system/machinecheck/machinecheck0/bank* # mcelog?
//      
// ToDo: drop cashes before start, lock mempage (to shm-file?)
// ToDo: spinpack: Immunity-aware programming, p.e. check sum matrix_elements
//     and compare every iteration (stored=memerror, recompute=compute_error?)
//
// before that do: echo 3 >/proc/sys/vm/drop_caches
//  cat /sys/devices/system/edac/mc/mc{0,1}/csrow{0,1}/{c,u}e_count|xargs
// use: dmidecode # to detect ECC-functionality
// GOMP_CPU_AFFINITY="0-15:1" 
// 2015-09x31_______2GB_rw1.5GB                 348MB/s 1t 2004 noECC PentiumM-1400MHz@1200MHz gcc451
// 2018-06comp1_4x2c32GB_rw3.6GB/c   -      -  1051MB/s 1t 2006 ECC-1defDIMM Opt885-2.6GHz+16*2GB-DDR-400MHz-gcc410vec read=2.3GB/s taskset+numctl ~openmp
// 2018-06comp1_4x2c32GB_rw3.6GB/c 5.4GB/s 8t  1055MB/s 1t 2006 ECC-1defDIMM Opt885-2.6GHz+16*2GB-DDR-400MHz-gcc492vec read8t=15.1GB/s read1t=2.3GB/s
//        comp1.taskset.01=10=1051MB/s 02=20=1064MB/s 04=40=1102MB/s 08=1061MB/s gcc410vec Board4.DIMM5=defekt
//        comp1.taskset.01=10=1046MB/s 02=20=1061MB/s 04=40=1099MB/s 08=1055MB/s gcc492vec_meggie
// 2015-09ibioN_2x2x1GB_rw900MB/c  2.8GB/s 4t  1030MB/s 1t 2006 ECC Opt2214-2.2GHz gcc412-O2+unroll+vect dmidecode=64bit-DIMMs board=SbECC (aber 9chips/DIMM) 2*2of4DIMMs gcc492+k8=2.5GB/s/4c bios.ecc=on memtest86+v420=2GB/s 4*singletask=4*720MB/s
// 2015-09ibioM_1x4x1GB_rw1.6GB/c  1.4GB/s 2t  1168MB/s 1t 2006 ECC Opt275-2.2GHz  gcc412-O2+unroll+vect dmidecode=72bit-DIMMs
// 2015-09mkautz12x2x2GB_rw500MB/c 876MB/s 6t   162MB/s 1t 2008 ECC m5k.gcc412 -O2 unroll no_openmp! 6t=6*single DDR2
// 2015-10mkautz12x2x2GB_rw500MB/c 6xxMB/s 6t   162MB/s 1t 2008 ECC m5k.gcc424 -O2 unroll +openmp t6schwankt=413-690MB/s! 2t=303MB/s 3t=420MB/s 4t=bad=290-305
// 2015-10mkautz12x2x2GB_rw500MB/c 780MB/s 6t   162MB/s 1t 2008 ECC m5k.pathcc -O2 unroll +openmp loop-var-ijk-must-be-signed!       3t=409MB/s 500MHz
// 2015-10asgard12x2x2GB_rw500MB/c 929MB/s 6t   203MB/s 1t 2008 ECC m5k.pathcc -O2 unroll +openmp                                    3t=505MB/s 680MHz
// 2015-10asgard12x2x2GB_rw500MB/c      -       202MB/s 1t 2008 ECC m5k.gcc412=gcc464 -O2 unroll 680MHz
// 2015-10asgard12x2x2GB_rw500MB/c 1.0GB/s 6t   198MB/s 1t 2008 ECC m5k.gcc464 -O2 unroll +openmp 680MHz mit/ohne bootstrap          3t=569MB/s 6t=6*1t-20%
// 2015-10asgard12x2x2GB_rw500MB/c 899GB/s 6t   170MB/s 1t 2008 ECC m5k.gcc492 -O2 unroll +openmp 680MHz --disable-bootstrap
// 2017-08kautz972x2x2GB_rw2000MB       -       140MB/s 1t 2009 ECC m5k.gcc412 -O2 500MHz DDR2-RAM
// 2017-08kautz972x2x2GB_rw400MB/c 701MB/s 6t   163MB/s 1t 2009 ECC m5k.pathcc -O2 500MHz +openmp
// 2016-09asgardsp_1x1GB_rw750MB/c      -  1t   874MB/s 1t 2008 Sempron3600+800MHz+DDR2-266MHz gcc412
// 2015-10mkautzsp_1x1GB_rw500MB/c      -       902MB/s 1t 2008 noECC?mc! Sempron3500+800MHz k8=sse2+3dnow
// 2015-09quantum1x4x2GB_rw1GB/c:  1.7GB/s 4t   1.2GB/s 1t 2008 ECC E5405-2GHz 228W-200W dmidecode=72bit-DIMMs DDR2 Multi-bit-ECC
// 2015-09quantum2x4x4GB_rw2GB/c:  2.3GB/s 8t   643MB/s 1t 2008 ECC E5450-3GHz
// 2015-09jsPC_____1x2GB_rw1GB     1.2GB/s 2t   1.2GB/s 1t 2010 noECC E5400-2.7GHz@1.6GHz
// 2016-06jsPC_____2x2GB_rw3GB    1.33GB/s 2t  1.30GB/s 1t 2010 noECC E5400-2.7GHz@1.6GHz
// 2015-09einstein2x2GB_rw1.5GB    2.7GB/s 4t   2.4GB/s 1t 2010 ECC? 4c8t-i7-860-2.8GHz gcc44-32bit! no-sse4-used! dmidecode=72bit-DIMMs BoardECC=none
// 2015-09landau__6x4GB_rw12GB     5.3GB/s 4t   4.1GB/s 1t 2010 ECC? 4c*2HT-iCore_i7-950-3.06GHz-8MB SSE4.2 24GB dmidecode=2304/2244bits? board=noECC
// 2015-08comp2__4x4x4GB_rw3GB/c: 15.8GB/s 16t  2.0GB/s 1t 2011 ECC Opt6134-2.3GHz
// 2015-09comp2n_4x4x8GB_rw7GB/c: 16.5GB/s 16t  2.1GB/s 1t 2011 ECC Opt6134-2.3GHz.gcc492-O2+unroll+vect
// 2015-09gibbs____4x4GB_rw14GB    5.9GB/s 4t   5.1GB/s 1t 2011 noECC? i5-2400-3.1GHz gcc412-O2+unroll+vect+omp hdc=30130h=3.4y dmidecode=64bit-DIMMs board=noECC
// 2015-09hpmini_2GB_rw1500MB:    1020MB/s 2t   990MB/s 1t 2012 noECC N455-1660MHz+gcc472-O2+unroll+vec BitError!
// 2015-09meitner__4x8GB_rw16GB    5.6GB/s 4t   5.3GB/s 1t 2013 noECC? 4c8ht-i7-3770-Gen2-3410MHz-avx DDR3 dmidecode=64bit-DIMMs
// 2015-09dyson__1x4x8GB_rw7GB/c:  6.8GB/s 4t   6.0GB/s 1t 2014 noECC i5-4460-3.2GHz+gcc482-25.6GB/s 80W dmidecode=64bit-DIMMs
// 2015-09dyson__1x4x8GB_rw7GB/c:  5.9GB/s 4t   2.1GB/s 1t 2014 noECC i5-4460-800MHz+gcc482 49W
// 2015-09msiKS____2x4GB_rw4GB     5.6GB/s 4t   5.1GB/s 1t 2015 noECC i5-4210H-2.9GHz+gcc492-O3+unroll+vec
// 2015-12t100 2x(E5-2630v3-2.4GHz-8c) 2x(2DIMMs/ch 1866MHz*8B=14.9GB/s/ch 4ch=59GB/socket)
// 2015-12t100_2x8x16GB_rw14GB/c: 30.4GB/s 16t  5.1GB/s 1t 2016 ECC 2xE5-2630-v3-2.40GHz+gcc483 # O2=30.4GB/s O2+unroll+vect=30.5GB/s
// 2015-12t100_2x8x16GB_rw14GB/c: 30.3GB/s 16t  5.xGB/s 1t 2016 ECC 2xE5-2630-v3-2.40GHz+gcc482 O2+unroll+vect=30.3GB/s n151.2016-12-12
// 2015-12t100_2x8x16GB_rw11GB/c: 30.7GB/s 16t  5.2GB/s 1t 2016 ECC 2xE5-2630-v3-2.40GHz+gcc482 O2+unroll+vect=30.5GB/s n151.2016-12-12
// 2015-12t100_2x6x16GB_rw11GB/c: 22.8GB/s 16t  5.0GB/s 1t 2016 ECC 2xE5-2630-v3-2.40GHz+gcc482 O2+unroll+vect=30.5GB/s n152.2016-12-12
// 2015-12t100_2x8x16GB_rw11GB/c: 30.3GB/s 16t  5.xGB/s 1t 2016 ECC 2xE5-2630-v3-2.40GHz+gcc482 O2+unroll+vect=30.5GB/s n152.2016-12-15 nach replace bad DIMM
// 2018-03t100_32*8GB_rw7GiB/c:   29.7GB/s 32t  4.9GB/s 1t 2016 ECC gcc483.O2=20GB/s,r=63GB/s 16t=30.4GB/s    32*7=224GiB=229376MB
// 2018-03t100_32*8GB_rw5.5GiB/c: 29.7GB/s 32t  4.9GB/s 1t 2016 ECC gcc483.O2=20GB/s,r=63GB/s 16t=30.4GB/s  32*5.5=176GiB=180224MB
// 2018-03t100_16*16GB_rw11GiB/c: 30.2GB/s 16t  5.0GB/s 1t 2016 ECC gcc483.O2=20GB/s,r=63GB/s 16t=30.4GB/s  32*5.5=176GiB=180224MB
// 2020-06t100_16*16GB_r8GB/c_omp: 92.2GB/s 8t  # auto-interleaved 8t? 16t!
// 2020-06t100_16*16GB_r8GB/c_omp: 13.7GB/s 8t  # 1-way/ch 1-way/csrow
//   until 2020-05 here to slow measurement
// 2020-06jsPC16_2*8GB_r1GB/c:    26.4GB/s  8t 15.7GB/s 1t 78%*34GB/s 2DIMMs*8GB-DDR4-2133MT/s=17GB/s 2ch=34.1GB/s i7-6700-3.4GHz-4c8t w=50%r ab 1-4GB
// 2020-06t100_16*16GB_r15GB/c    48.2GB/s  8t 11.5GB/s 1t # cpu0 m0 only w=22.6GB/s ab30GB 1cpu8c=59GB/s*82%
// 2020-06t100_16*16GB_r15GB/c     7.8GB/s  8t  7.5GB/s 1t # cpu0 m1 only w=5.0GB/s 1cpu8c=59GB/s_max
// 2020-06t100_16*16GB_r15GB/c    95.3GB/s 16t 11.5GB/s 1t # 2cpus/node w=44.2GB/s ab30GB 1cpu8c=59GB/s*81% no-vec
// t100: GOMP_CPU_AFFINITY="0-15" srun -p sw01_short -n 1 -t 00:10:00 bash -c
//        "OMP_NUM_THREADS=16 ./mem_badram 245760" # (16c*14GB/c=224GB)*1024MB/GB
//        # 2018-03 change to 0-31 !?
//  OMP_NUM_THREADS=16;GOMP_CPU_AFFINITY="0-15" numactl -l -C 0-15 ./mem_badram $[240*1024] # 30.5GB/s
//  OMP_NUM_THREADS=8;GOMP_CPU_AFFINITY="0-7"   numactl -l -C 0-7  ./mem_badram $[120*1024] # 15.3GB/s
//  OMP_NUM_THREADS=8;GOMP_CPU_AFFINITY="0-7" numactl -m 1 -C 0-7  ./mem_badram $[120*1024] #  3.0GB/s
//    check correct pinning on running HPC (high-mem) apps (health?)
//     + other 8 cores crosswise: 2*3064MB/s  check: ps -am -o spid,psr,cmd
//    test -N 3 OK, c129 = 9GB/s only (1st round, drop_caches?), c129.2nd = 22.5GB/s
//     3th = 24.6GB/s  ToDo: min/max nodes/cores
// cmp: https://www.techspot.com/news/62129-ddr3-vs-ddr4-raw-bandwidth-numbers.html
//   AIDA64 Read Bandwith? aus Bild (cpu/memctlr-abhaengig!)
//   read-DDR3-1866-4ch = 49GB/s = 59.7GB/s * 82%  write=33GB/s=67%r cp=43GB/s
//   read-DDR4-2133-4ch = 54GB/s = 68.3GB/s * 79%  write=47GB/s=87%r cp=52GB/s
//  DDR3 1600/1866/2133/2400/2666/2800MHz  t100=1866MHz(4*14.9GHz/cpu)
//  DDR4 2133/2400/2666/2800/3000/3200MHz  PC16=2133MHz(2*17GB/s*78%)
//    latencies 45(2800MHz)-80ns(1600MHz)
//
/* --- 2015-09hpmini_2GB_2012 bit-flip-error ---
 malloc 1024 MiB (Mi=2^20) long= 8 B = 64 bit (arg1)
 malloc 1024 blocks to 1024 kiB = 131072 * long (ki=2^10)
 starting, only every 128th block shown ...
 a=00007fb978db0010 = 130789 GB blk    0 (alloc+ini)
 a=00007fb9708a0010 = 130789 GB blk  128 (alloc+ini)
  ...
 a=00007fb948620010 = 130789 GB blk  768 (alloc+ini)
 a=00007fb9405a0010 = 130789 GB blk  896 (alloc+ini)
 a=00007fb978db0010 iLoop=00/16 ffffffffffffffff
 a=00007fb978db0010 iLoop=02/16 ffffffffffffffff
 loop=  0 speed(w+r):        887 MB/s (1500.000 MB)
 a=00007fb957361010 iLoop=00/16 9bc3679699ccb78c
 a=00007fb957361010 iLoop=02/16 9bc3679699ccb78c
 a=00007fb957361010 iLoop=04/16 9bc3679699ccb78c  # 1-to-0 bit 15
ERR bad data blk=734 lrnd= 9bc3679699ccb78c  sum=00008000 (j=4) on ?
ERR bad data   a[0001c603]=9bc3679699cc378c diff=00008000 (j=4)
ERR bad data blk=734 lrnd= d5176bb59daaa713  sum=00008000 (j=2) on ?
ERR bad data   a[0001c603]=d5176bb59daa2713 diff=00008000 (j=2)
ERR bad data blk=734 lrnd= ffffffffffffffff  sum=00008000 (j=8) on ?
ERR bad data   a[0001c603]=ffffffffffff7fff diff=00008000 (j=8)
ERR bad data blk=734 lrnd= ee6f32189910bc72  sum=00008000 (j=10) on ?
ERR bad data   a[0001c603]=ee6f321899103c72 diff=00008000 (j=10)
ERR bad data blk=734 lrnd= ef6306d1f2128d9c  sum=00008000 (j=2) on ?
ERR bad data   a[0001c603]=ef6306d1f2120d9c diff=00008000 (j=2)
ERR bad data blk=734 lrnd= 7fdcc2331befd79f  sum=00008000 (j=1) on ?
ERR bad data   a[0001c603]=7fdcc2331bef579f diff=00008000 (j=1)
ERR bad data blk=734 lrnd= be583b369497f865  sum=00008000 (j=2) on ?
ERR bad data   a[0001c603]=be583b3694977865 diff=00008000 (j=2)
 haeufig, dann weniger, dann haeufiger bei parallelem 20MB-scan = 20kB blks
 heating-related? ToDo: how to lock? as /dev/shm-file?
 ext2 did crash spontaniosly, so I did the memory test
   (bad) mem is used as dsk-cache, good for speed, bad for data integrity
   ToDo: fuse-project crc-fs for ramdisks (tiny-core)
 addr changing from time to time
 memtest: DDR2-333 2GB 2GB/s
   error: 00023e8d023 - 574.5 MB  2015-09-09 kernel badram/badmem-fix?
  locking survives s2ram + s2dsk = suspend-to-disk = hibernate (2 tests)
 kautz18: edac-ce+ue-errors see r_kautz.txt
   pathcc -O2 -fopenmp -D_OPENMP -static -o mem_badram mem_badram.c
   x=$(date +%F_%T);srun -N1-980 -l -W 0 -p big --time=0-9 --cpus-per-task=6 \
      --task-mem=3000 ./mem_badram 2400 -l100;echo $x;date +%F_%T # -l2=2min
 18.03.18 So 07:40 health + srun -N900-970 -l -W0 ... 
  edac-err scx-m10n8  0 0 0 0 0 0 0 0 1 0 0 0           # cpu0/ce
  edac-err scx-m19n11 0 0 0 0 0 0 0 0 481637 0 105584 0 # cpu0/ce dma0/ce
  edac-err scx-m34n2  0 0 0 0 0 0 0 0 24698 0 0 0
  edac-err scx-m34n4  0 0 0 0 0 0 0 0 4 0 0 0
  edac-err scx-m34n7  0 0 0 0 0 0 0 0 3671 0 0 0
  edac-err scx-m34n15 0 0 0 0 0 0 0 0 39478 0 126049 0
  +m34n0 +m34n12
  m34n15: OMP_NUM_THREADS=6 ./mem_badram 2400 -l2;date +%F_%T # 2min
  m19n11: OMP_NUM_THREADS=1 ./mem_badram 2400 -l2;date +%F_%T # 
  m34n12: 
 using OpenMP OMP_NUM_THREADS= 6
edac    0=scx-m34n12.scsystem
edac    0@00 /sys/devices/system/edac/cpu/cpu0/ce_count   :      6      +6
 malloc 2400 MiB (Mi=2^20) long= 8 B = 64 bit (arg1)
 malloc 1024 blocks to 2400 kiB = 307200 * long (ki=2^10)
 starting, only every 128th block shown ...
 a=0000005569570010 = 341GB blk    0 sz=2400kB (alloc+ini)
edac    0@00 /sys/devices/system/edac/cpu/cpu0/ce_count   :     15      +9
 a=0000005569570010 iLoop=00/16 ffffffffffffffff
ERR blk= 539 bad data lrnd= ffffffffffffffff  sum=0000120000000000 j=0
ERR bad data   a[00008e86]= ffffedffffffffff diff=0000120000000000 j=0 1
ERR blk= 539 range=00008e86 sz=00000001 cnt=1 1
edac    0=scx-m34n12.scsystem
edac    0@00 /sys/devices/system/edac/cpu/cpu0/ce_count   :     36     +21
 loop=  0 speed(w+r):      12806 MB/s (2400.000 MB)
...
# 2nd run 6threads # other address?
 a=00000055697b0010 = 341GB blk    0 sz=2400kB (alloc+ini)
  ...
ERR blk= 539 bad data lrnd= ffffffffffffffff  sum=0000120000000000 j=0
ERR bad data   a[00014e86]= ffffedffffffffff diff=0000120000000000 j=0 1
ERR blk= 539 range=00014e86 sz=00000001 cnt=1 1
edac    0@00 /sys/devices/system/edac/cpu/cpu0/ce_count   :     49      +4
 loop=  0 speed(w+r):      12821 MB/s (2400.000 MB)
 ERR blk= 539 bad data lrnd= ffffffffffffffff  sum=0000120000000000 j=32
 ERR bad data   a[00014e86]= ffffedffffffffff diff=0000120000000000 j=32 1
 ERR blk= 539 range=00014e86 sz=00000001 cnt=1 2
 loop=  0 speed(r):         1836 MB/s (2400.000 MB)
 loop=  0 speed(rh):        2851 MB/s (2400.000 KB)
 ERR blk= 539 bad data lrnd= ffffffffffffffff  sum=0000120000000000 j=32
 ERR bad data   a[00014e86]= ffffedffffffffff diff=0000120000000000 j=32 1
 ERR blk= 539 range=00014e86 sz=00000001 cnt=1 3
 a=00000055697b0010 iLoop=00/16 9bc3679699ccb78c
edac    0@00 /sys/devices/system/edac/cpu/cpu0/ce_count   :     58      +9
 a=00000055697b0010 iLoop=02/16 9bc3679699ccb78c
 a=00000055697b0010 iLoop=04/16 9bc3679699ccb78c
 a=00000055697b0010 iLoop=08/16 9bc3679699ccb78c
 loop=  1 speed(w+r):        809 MB/s (2400.000 MB)
 loop=  1 speed(r):         1862 MB/s (2400.000 MB)
 loop=  1 speed(rh):        2849 MB/s (2400.000 KB)
1 bad blocks detected at task0=scx-m34n12.scsystem, start sleeping ...
for x in $(find /sys/devices/system/edac -name \*e_count);do echo $x $(cat $x);done
/sys/devices/system/edac/cpu/cpu0/L2/ue_count 25
/sys/devices/system/edac/cpu/cpu0/L2/ce_count 58
/sys/devices/system/edac/cpu/cpu0/ue_count 25
/sys/devices/system/edac/cpu/cpu0/ce_count 58   # ue!!! but RAM, not L2
# vs=401MB rss=6MB pid=6374 free=3.4GB
cat /proc/6374/smaps 
55b97d0000-55b9a30000 rw-p 55b97d0000 00:00 0 
Size:              2432 kB
Rss:               2432 kB
Private_Dirty:     2432 kB
cat /proc/6374/maps
55b97d0000-55b9a30000 rw-p 55b97d0000 00:00 0 
dmesg: EDAC DEVICE1: UE: cpu instance: cpu0 block: L2 'L2 Data ECC Error: Address=0x93637bc0'
# addresses changing max_ca 0x934f70c0 1<<31=2GB, 0x13a7c0=2^20=1MB
m34n12 during sleeping 2nd run
m34n12: x=$(date); ./mem_badram 2400 -l2;echo $x;date # -l2=2min
edac    0@00 /sys/devices/system/edac/cpu/cpu0/ce_count   :     58
edac    0@00 /sys/devices/system/edac/cpu/cpu0/ue_count   :     25

2020-01 bb.urz 2 DIMMs DDR 400MHz ECC mc0.csrow6 ce=4000/d+ue=145/225d
   errors on every loop, every bitmap (bios.dimm.interleaved?)
*/      
//  https://en.wikipedia.org/wiki/ECC_memory
//    spacecraft Cassini–Huygens 1997 2.5Gbit-ECC 280-single-bit-errors/day
//               on day 1997-11-06 4xErrorRate (solar-event)
//    pubs 2007-2009 1e-10..1e-17 errors/bit*h = ca. 1err/GB*h or less
//    google: 8% od DIMMs per year, 7e-11err/bit*h or 5err/8GB*h Streuung?
// mem-intense apps may be more efficient on low-clock (or add cpu-intense app, pe. thermo+psu-stress)
// ToDo: add volatile array, correct speed (less cache dependency)
//  gcc492 -O2 -fopenmp -fverbose-asm -Wa,-adhln -g mem_badram.c
//   ASM? gcc492: unroll8*xmm0(128bit).store + unroll8*rax.cmp
// ToDo: does suspend to disk, reorders phys memory? locking lost?
// ToDo: if found, test time to flip or row-hammer dependence (idle)
//
// bad-mem-successed on:
//  yyyy-mm-dd machine remarks


#include <unistd.h>  /* sleep(60) needs extra h-file */
#include <stdlib.h>  /* system("sleep 600"); */
#include <stdio.h>
#include <sys/time.h>
#include <string.h>   /* strncmp() */
#ifdef _OPENMP
#include <omp.h>
#endif
#include <sys/mman.h>   /* Memory locking functions */

typedef unsigned long int ULong;

double fstime(){
 struct timeval tv; if(gettimeofday(&tv, NULL))perror("fstime");
 return tv.tv_sec+1e-6*tv.tv_usec; }
 
#define INVLOOPs 16  /* 2015-08, if 1 gcc492 optimizes loops away ToDo: ??? */

char * dummy_nodename = "?";
char * slurmd_nodename = "??"; /* SLURMD_NODENAME */
int edac_errors = 0;   /* ... since program start (not the old once) */
/* bit0= invers, bit1= .x.x.x.x, bit2=..xx..xx, bit3=....xxxx bit10=last?  */
int edac_blk_bitmap = 0; // 1<<n for    .x.x.x ..xx..xx ....xxxx  0=init
int edac_blk_invers = 0; // 0 or ~0 for x.x.x. xx..xx.. xxxx....   
int edac_blk_idx = 0 ; // errors triggered by bitmap filter    bitmap>>1
#define NUM_EDAC (4+4+4+4+3)  // max. 31 (32bit)
/*
 for x in $(find /sys/devices/system/edac/ -name \*panic_on_ue);do\
  y=$(cat $x);if [ $y != 0 ];then echo $x $(cat $x);fi;done
 for x in $(find /sys/devices/system/edac/ -name \*e_count);do\
  y=$(cat $x);if [ $y != 0 ];then echo $x $(cat $x);fi;done
 # panic_on_ue kautz=yes        t100=no
 # dma0        kautz=yes(ram?)  t100=no
 # cpu0        kautz=RAM_errors t100=no
 # mc0 mc1     kautz=allways0   t100=yes
 # /sys/module/edac_core/parameters/edac_mc_panic_on_ue  kautz=0 t100=0
 # /sys/module/edac_core/parameters/edac_pci_panic_on_pe kautz=0 t100=0
 # /sys/kernel/debug/mce/fake_panic                     kautz=no t100=0
 #  2020-03-26 1t00 trotzdem paniced = reset + B1.double-bit-error displayed
 # 
*/  
char *edacs[NUM_EDAC]={  // better dynnamic?
    "/sys/devices/system/edac/mc/mc0/ce_count",
    "/sys/devices/system/edac/mc/mc0/ce_noinfo_count",
    "/sys/devices/system/edac/mc/mc0/ue_count",
    "/sys/devices/system/edac/mc/mc0/ue_noinfo_count",
    "/sys/devices/system/edac/mc/mc1/ce_count",
    "/sys/devices/system/edac/mc/mc1/ce_noinfo_count", /* 39 chars */
    "/sys/devices/system/edac/mc/mc1/ue_count",
    "/sys/devices/system/edac/mc/mc1/ue_noinfo_count",
    "/sys/devices/system/edac/mc/mc2/ce_count",     // comp2 mc0..mc3 180414
    "/sys/devices/system/edac/mc/mc2/ce_noinfo_count",
    "/sys/devices/system/edac/mc/mc2/ue_count",
    "/sys/devices/system/edac/mc/mc2/ue_noinfo_count",
    "/sys/devices/system/edac/mc/mc3/ce_count",
    "/sys/devices/system/edac/mc/mc3/ce_noinfo_count",
    "/sys/devices/system/edac/mc/mc3/ue_count",
    "/sys/devices/system/edac/mc/mc3/ue_noinfo_count",
    "/sys/devices/system/edac/cpu/cpu0/ce_count",  // kautz
    "/sys/devices/system/edac/cpu/cpu0/ue_count",  // kautz panic_on_ue==0?
    "/sys/devices/system/edac/dma/dma0/ce_count"};
// /sys/devices/system/node/node[01]/numastat t100.TODO numa_hit

int chk_edacs=(1<<NUM_EDAC)-1; /* bit pattern? bit/path_in_list_above? */

// analyse memory given by pointer and length in sizeoflong
//  gcc492 -O2 -funroll-loops ... ToDo: optimiert loops to cache?
//    num = mb*(1024l/sizeof(ULong));  /* memory in kB/sizeLong */
//    v2015-08: *a=512MB to 512kB-Blocks/core (L2-cache) = Bad (ToDo **a)
//  ToDo: multiple copy mini-array(4kB)
// use volatile to avoid bad optimization, we expect possible memory change
//   problem volatile versus vectorization? check asm-code!
//   gcc492 -O2 -fopenmp -fverbose-asm -Wa,-adhln -g mem_badram.c | less # ASM-code
// all blocks in INV-loop to avoid in-cache work
// old 2017-- write all (200GB/100s), read all (200GB/100s)
//      100s active measure time between write and next read,
//      but +100s useless time between read and next write
// ToDo: better write new pattern to block after read+check a block
//      1/1000 useless time + double measure time (option: initial = skip check)
//      also writing in neighbourhood of reading is kind of hammering?
//      measure time expandable by SIGSTOP

// added 2018-03 to get ECC response, j=0..INVLOOP-1
int chk_edac(int procid, int ii, int j, ULong *bad, int *bad_nb) { // j=LOOP.j
  FILE *p1=NULL;
  // added 2017-08 Todo: skipmodus on err
  int i1, rc=0, i2[NUM_EDAC]={0,0,0,0,0,0,0,0,0,0,0};
  static int    i3[NUM_EDAC]={0,0,0,0,0,0,0,0,0,0,0};
  static int out_nodename=1; /* output once only on first error */
  if (chk_edacs) {
    if (edac_errors) usleep(1000000); // added 1s 2018-04-04, edac seems to need relax time
    // without that time binary tree bad block search does work badly 
    //  ... finds different blocks, maybe because edac delay
    for (i1=0;i1<NUM_EDAC;i1++)
    if (1 & (chk_edacs >> i1)) { // loop over path list 
      p1=fopen(edacs[i1],"r");
      if (!p1) { // suppress 2nd failed-message 2020-06
        if (procid==0 &&  i1>=1 && 1==(1&((chk_edacs<<1)>>i1)) )
          printf("chk_edac.open %-47s ok ...\n",edacs[i1-1]);
        if (procid==0 && (i1==0 || 1==(1&((chk_edacs<<1)>>i1))))
          printf("chk_edac.open %-47s failed ...\n",edacs[i1]);
        chk_edacs &= ~(1<<i1); } // added 2018-03 for non-ECC
      if (p1 && fscanf(p1,"%d",&i2[i1])!=1)
        printf("chk_edac.read %-47s failed ...\n",edacs[i1]);
      if (p1 && fclose(p1)) perror("err fclose");
      if (i2[i1]!=i3[i1]) {
        printf("edac %4d i=%4d.%02d %-45s %5d %+3d\n",
           procid, ii, j, edacs[i1], i2[i1], i2[i1]-i3[i1]);
        rc += (i2[i1]-i3[i1]); // return new errors
        i3[i1]=i2[i1];
      }
      // if (i1>1 && chk_edacs==0) break;
    }
  }  // chk_edacs,
  //  ### popen gives no err if cat failes
  //p1=popen("cat /sys/devices/system/edac/mc/mc[01]/[cu]e_{,noinfo_}count"
  //   " /sys/devices/system/edac/cpu/cpu[0-5]/[cu]e_count|xargs 2>&1","r");
  // for (rc=i1=0;i1<NUM_EDAC;i1++) rc += i2[i1];
  if (rc && out_nodename) { out_nodename=0; /* output once only */
    printf("edac %4d=%s rc=%d\n", procid, slurmd_nodename, rc);
  }
  // if no error switch blk_inverse, on error set bit, next bit if bit found
  if (rc && edac_blk_bitmap) {
    // 2018-04-04 add initial blk_idx for new cycle
    //   ToDo: after 2 detects, mark blk as bad and go around!?
    if (edac_blk_bitmap==1) { edac_blk_idx=0; }
    // bitmap & 1 == 0 for .x.x.x, == 1 for x.x.x.
    if (!edac_blk_invers) edac_blk_idx |= edac_blk_bitmap;
    else if (edac_blk_idx & edac_blk_bitmap)
      printf("ERR edac_blk_bitmap (more than one bad block?)\n");
    printf("edac %4d=%s loop=%4d.%02d rc=%d inv=%d bitmap=%04x"
           "  blk_idx=%04x %4d\n",
          procid, slurmd_nodename, ii, /* invloop */ j, /* ecc-errors */ rc, 
          edac_blk_invers&1, edac_blk_bitmap,
          edac_blk_idx, edac_errors);
    if (edac_blk_bitmap==1<<9) { /* final of 1024 blocks */
      static int last_blk=0, repeat=0;
      if (last_blk==edac_blk_idx) { repeat++;
        if (repeat>2) { bad[last_blk]++; bad_nb++;
          printf("WARNING: %3d mark block %4d 0x%04x as bad, continue ...\n",
                 procid, last_blk, last_blk); repeat=0; last_blk=0;
        } /* mark blk as bad + continue (further blocks?) */
      } else
      if (repeat==0) { last_blk=edac_blk_idx; repeat++; } else
      { /* block changing */ }
      
    }
  } 
  return(rc);
} // end edac-pipe-module
  

int mem_wr(ULong volatile ** pa,long nb, long num, ULong lrnd,
           int procid, ULong *bad){
  long i, k;
  ULong volatile * a;
#pragma omp parallel for default(shared) private(i,k,a) firstprivate(lrnd,nb,num) schedule(static)
  for (i=0;i<nb;i++) { /* 1024 blocks writing */
    a = pa[i];
    if (bad[i]) continue; // 2018-03 skip bad blks, find others too
    // bitmap & 1 == 0 for .x.x.x, == 1 for x.x.x.
    if (pa[i] && (((i^edac_blk_invers)&edac_blk_bitmap) || !edac_blk_bitmap)) {
      // ToDo: vectorize lrnd, do not write all every loop (rowhammer)
//#pragma omp parallel for default(shared) private(k) firstprivate(lrnd) schedule(static)
      for (k=0; k<num; k++) a[k]=lrnd; // 2015-08-31 rnd changed to lrnd
      // ToDo: fast scan for error + if error slow scan + show info
      // ToDo: ompi-reduction!
    } // if valid block
  } // block-loop (big data avoids cache intern work) 2015-09-03 2nb-loops
  return 0;
}

// t100.n146 86-93GB/s=89.5+-4%, why not const?  2019-01
int mem_rd(ULong volatile ** pa,
    const long nb, const long num, ULong lrnd,
    const int procid, ULong *bad, int j){ // j = round index
  long j1=0, j2=0, /* addr range */ jn=0, /* num */ i, k, sum=0;
  int bad_nb=0; /* bad blks counter (bad_NumBlocks) */
  static long jn2=0;
  ULong volatile * a;
#pragma omp parallel for default(shared) private(i,k,a) firstprivate(lrnd,num,nb) reduction(|:sum) schedule(static)
  for (i=0;i<nb;i++) { /* 1024 blocks reading */
    a = pa[i];
    // 2018-04-03 try to fix bad data ffffff after EDAC error
    if (pa[i]) if (pa[i][0]==~lrnd) lrnd=~lrnd; // 2018-04-03 bad fix
    if (bad[i]) continue; // 2018-03 skip bad blks, find others too
    if (pa[i] && (((i^edac_blk_invers)&edac_blk_bitmap) || !edac_blk_bitmap)) {
      // ToDo: vectorize lrnd
      sum=0; j1=j2=jn=0;
// TODO parallelize outer loop i=0..1023  asm L47 unroll8*xorq no-vec gcc8
//   no speedup with sum1+2 lrnd1+2 k+=2 ... k+=4
//  TODO 2modes: parallel k or parallel i (test speed non-interleaved!)
//#pragma omp parallel for default(shared) private(k) firstprivate(lrnd,num) reduction(|:sum) schedule(static)
// #pragma GCC unroll (8)
      for (k=0; k<num; k++) 
        sum |= a[k]^lrnd; // 2015-09-03 vectorize +5%
      /* 2nd slow scan on error! possible it reads from CPU cache now? */
      if (sum) { // bit error handling (wegoptimiert!?)
        if (!(bad_nb & (bad_nb-1))) // jn2 is bad_byte_counter, better use bad_blk_count
          printf("ERR blk=%4lu adr=%16p bad data on thread %d %s bad_nb=%d\n",
                          i, a, procid, slurmd_nodename, bad_nb+1);
        if (jn2==1024) printf("ERR suppress further messages ...\n"); // 2020
        if (jn2<1024 || (jn2 & (jn2-1))==0)
          printf("ERR blk=%4lu bad data lrnd= %016lx  sum=%016lx j=%u\n",
            i, lrnd, sum, j /* round */);
        /* allows ^c on long scrolls */
        if (jn2>1023 && (jn2 & (jn2-1))==0) usleep(100000); /* 100ms */
        j1=j2=jn=0;
        for (k=0; k<num; k++) if (a[k]!=lrnd) { 
          jn++; jn2++; j2=k+1; /* j2>j1 means bad range set */
          if ((jn2<1024 && jn<16) || (jn & (jn-1))==0)
          printf("ERR bad data   a[%08lx]= %016lx diff=%016lx j=%u %lu\n",
            k, (ULong)a[k], (a[k]^lrnd), j /* round */, jn);
        } else { if (j1==j2) j1=j2=k+1; }
        if (j2==j1) printf("ERR blk=%4lu bad data could not localized\n",i);
        else
          if (jn2<1024 || (jn2 & (jn2-1))==0)
             printf("ERR blk=%4lu range=%08lx sz=%08lx cnt=%lu %lu\n",
                     i, j1, j2-j1, jn, jn2);
        if (jn2<1024 || (jn2 & (jn2-1))==0)
          printf("ERR reread to see permanence (caching %ldKB/cores?)\n",
            num/1024l);
        /* kautz 3GB/6c/1000=500kB vs. 256KB-L2
         * t100  250GB/16c/1000=15.6MB vs. 256KB-L2 */
        j1=j2=jn=0;
        for (k=0; k<num; k++) if (a[k]!=lrnd) { 
          jn++; j2=k+1; /* j2>j1 means bad range set */
          if ((jn2<1024 && jn<16) || (jn & (jn-1))==0)
          printf("ERR bad data   a[%08lx]= %016lx diff=%016lx j=%u %lu r\n",
            k, (ULong)a[k], (a[k]^lrnd), j /* round */, jn);
        } else { if (j1==j2) j1=j2=k+1; }
        if (j2==j1) printf("ERR blk=%4lu bad data could not localized\n",i);
        else
          if (jn2<1024 || (jn2 & (jn2-1))==0)
             printf("ERR blk=%4lu range=%08lx sz=%08lx cnt=%lu %lu reread\n",
                     i, j1, j2-j1, jn, jn2);
      }
    } // if good block (no errors until yet)
    if (jn/* || sum*/) { bad[i]++; bad_nb++; jn=0; } // count bad blocks + mark as bad
  }  // 1024 mem-blocks i=...
  return bad_nb;
} // mem_rd

/* read 50% of 32 Blocks=1.6% must be greater than cache */
// read pattern: x.x.x.x.  x.x.... 
int mem_rd_hammer(ULong volatile ** pa,long nb, long num, ULong lrnd,
           int procid, ULong *bad, int j){ // j = round index
  long jj=0, /* i,*/ k, sum=0;
  int i = lrnd % nb; // random block
  ULong volatile * a = pa[i];
  if (nb<128)  { printf("WARN not enough blocks\n"); exit(0); }
  if (pa[i] && (((i^edac_blk_invers)&edac_blk_bitmap) || !edac_blk_bitmap)) {
    // ToDo: vectorize lrnd
#pragma omp parallel for default(shared) private(jj,k,i)\
  firstprivate(lrnd) reduction(|:sum) schedule(static)
    for (jj=0;jj<1024*2;jj++)
     /* for (i=0+32;i<64+0*nb;i++) */ { /* 1024*32 blocks reading */
      for (k=0; k<num; k+=8) sum |= a[k]^lrnd; // 2015-09-03 vectorize +5%
      /* slow on error! */
    } // if good block (no errors until yet)
  }  // 1024 mem-blocks i=...
  return (sum==0);
} // mem_rd

// write + readchk (better read and check old + write for next?)
int chkmem(ULong volatile ** pa,  /* array of mem blocks, return: num errors */
    long nb, long num, /* 2D-array params nb=1024(or less on failure) */
    int ii, /* loop count 0..99 for compact debug output */
    ULong rnd, /* random 64bit-number */
    int procid, /* slurm_proc_id=0..n-1 */
    ULong *bad,  /* marker array [1024] */
    int sum_bad_nb) /* for dbg num_bad_blocks state, better outside chkmem? */
{
  double t1,t2,t3; // 2020-06
  ULong lrnd=rnd /* rnd-seed */; int i, j /* round-idx */, rc_bad_nb=0;
  int bad_nb=0, i_bad0=0, i_bad1=0; /* outputting from first_bad_blk to last_bad_blk */
  // char s2[1024]={0}; // buffer edac logs (see below)
  for (j=0;j<INVLOOPs; j++){ /* fix20150902: why j=1 ... ? bad measure! */
    lrnd=~lrnd;  /* inverse bitpattern between tests */
    /* j=0..7 ~rnd rnd ~rnd rnd ... ~rnd rnd */
    /* to see state of errors on screen at first sight even after 100 loops */
    if (procid==0 && ((j&(j-1))==0) /* 180414 log2 */
                  && ((ii&(ii-1))==0 || ii<8)){ /* log reduced output */
      bad_nb=0;
      for (i=0;i<1024;i++) { if (bad[i]) break; } if (i<1024) i_bad1=i_bad0=i;
      for (   ;i<1024;i++) if (bad[i]) { i_bad1=i; bad_nb++; }
      //printf(" a=%016lx iLoop=%4d.%02d/%d %016lx bad= %4d %4d nb %4d %4d\n",
      //  (ULong)pa[0], ii, j, INVLOOPs, lrnd,
      //   i_bad0, i_bad1 /* blk range */, bad_nb, sum_bad_nb+rc_bad_nb);
      printf(" iLoop=%4d.%02d/%d %016lx bad= %4d %4d nb %4d %4d\n",
        ii, j, INVLOOPs, lrnd,
        i_bad0, i_bad1 /* blk_range */, bad_nb, sum_bad_nb+rc_bad_nb);
/* ToDo: add output which block-range? show only log2(ii) or 0..3 + runtime */
      fflush(stdout); // slow?
      /* slurm needs flush to see progress 2015-09-03 */
    }
    t3=fstime();
                mem_wr(pa, nb, num, lrnd, procid, bad);
    // ToDo additional variing sleep here?
    t1=fstime();
    rc_bad_nb+= mem_rd(pa, nb, num, lrnd, procid, bad, j);
    t2=fstime();
    // TODO half blocks on ecc-error search (edac_blk_bitmap)
    if (procid==0 && t1!=t2 && t1!=t3 && (ii&(ii-1))==0 && (j&(j-1))==0) {
      printf(" iLoop= %3d.%02d speed.w: %10.3f GiB/s (%.0f MiB)\n",
             (int)ii, (int)j, ((edac_blk_bitmap)?0.5:1.0)*
              1.*nb/1024.*num*sizeof(pa[0])/1024./1024./(t1-t3), // = mb/t
                          num*sizeof(pa[0])/1024.*1.*nb/1024.);
      printf(" iLoop= %3d.%02d speed.r: %10.3f GiB/s (%.0f MiB)\n",
             (int)ii, (int)j, ((edac_blk_bitmap)?0.5:1.0)*
              1.*nb/1024.*num*sizeof(pa[0])/1024./1024./(t2-t1), // = mb/t
                          num*sizeof(pa[0])/1024.*1.*nb/1024.);
      fflush(stdout);
    } // print speed
    // echo 0 > /sys/devices/system/edac/mc/edac_mc_panic_on_ue
    // echo 0 > /sys/devices/system/edac/cpu/panic_on_ue
    // mkautz: /sys/devices/system/edac/mc/mc[01]/[cu]e_{,noinfo_}count
    //         /sys/devices/system/edac/cpu/cpu[0-5]/[cu]e_count # m34n15
    //         /sys/devices/system/edac/dma/dma0/[cu]e_count     # m34n15
    //         /sys/devices/system/edac/fsw/fsw0/ce_count
    //         /sys/devices/system/edac/dma/panic_on_ue # 1
    //         /sys/devices/system/edac/cpu/panic_on_ue # 1
    //         edac_mc_panic_on_ue # 1
    // mem_badram 1500 # runs, 2000 = segfault on 3GB free wegen pipe()=fork ?
    // create fpipe? + endlessloop cat_edac + sleep1  + read via fgets? 
    //   ToDo: non-root-access? t100 mc0 readable for all
    { int ce0, i; char xx[1024+1];
      ce0= chk_edac(procid,ii,j,bad,&rc_bad_nb); // ToDo: sleep 1s before read?
      if (ce0) edac_errors += ce0; // static var
      if (ce0) usleep(10000); // new errors
      // ToDo: if correctable-errors detected (after 1s) ... subblk-check
      //  try to find the bad memblock wr/rd .x.x.x x.x.x. ..xx..xx xx..xx
      if (edac_blk_bitmap) edac_blk_invers ^= ~0; // until ce0 found
      if (ce0) {
        if (edac_blk_bitmap==0) { edac_blk_bitmap=1; edac_blk_invers=0; }
        else {
           if (edac_blk_bitmap==1<<9) edac_blk_bitmap=1; else
           if (edac_blk_bitmap)       edac_blk_bitmap<<=1;
           edac_blk_invers=0;
        }
        for (i=0;i<33;i++)
          if (((i^edac_blk_invers)&edac_blk_bitmap) || !edac_blk_bitmap) 
            xx[i]='x'; else xx[i]='.';
        xx[i]='\0';
        if (edac_blk_bitmap<=32)
          printf("INFO test bitmap=%04x inv=%d  blk_pattern= %s\n",
           edac_blk_bitmap,edac_blk_invers&1,xx);
      } // edac-error found
    }  // chk_edac
    // j&1 leaves memory written with lrnd (not ~lrnd)
    //  difficult to detect case odd loop break and edac_blk_pattern written
    if (rc_bad_nb && (j&1)) break; // bad for rereads if invers written (even)
  } // INV-loop j=0..15
  
  return rc_bad_nb; /* return code is num bad_nb for lrnd 16 loops */
}

int main(int argn, char *argv[]){
  static ULong volatile * pa[1024]; /* no stack, segfault at getenv */
  ULong rnd=0, rnd1, rnd2; long ii, i, j;
  static ULong bad[1024]; /* no stack */
  long num, nb=0, mb=1024; // 2015-10 signed for pathcc openmp
  double t1, t2; 
  char *sprocid=NULL;
  int n_loops=100; /* number of loops (of INVLOOPs with different RND) */
  int procid=0, num_nodes=1; /* slurm_procid */
  int verbose=0, bad_nb=0; /* 2018-04-04 bad_nb changed from long to int */
//  int verbose=1; /* 0=quiet(for multinode) ToDo: check slurm_env! */
#ifdef _OPENMP
  int omp_n= 1;  /* openmp-threads */
#endif
  int sl=sizeof(ULong);            /* sizeLong should be 8 on 64bit or 4 on 32bit */
  // if (argn>1) mb=atoi(argv[1]);    /* memory in MB */
  for (i=1;i<argn;i++) {
    if (argv[i][0]>'0'  && argv[i][0]<='9') mb=atoi(argv[i]); /* MB */ else
    if (argv[i][0]=='-' && argv[i][1]=='l') n_loops=atoi(argv[i]+2); else
    if (argv[i][0]=='-' && argv[i][1]=='v') verbose=atoi(argv[i]+2); else
     { printf(
    "usage: mem_badram [<megabyte=1000>] [-l<nloops=100>] [-v<verbose=0>]\n");
      exit(1); }
  }
  system("date +%F\\ %T"); // 2018-04-14 for better output-documentation
  num = mb*(1024l/sizeof(ULong));  /* memory in kB/sizeLong */
  // SLURM_PROCID  #  0..SLURM_NPROCS
  if (verbose>6) printf("DBG L%d\n",__LINE__);
  sprocid=getenv("SLURM_PROCID");  /* string: 0..n or "" 2018-03 segflt */
  if (!sprocid) 
    sprocid=getenv("OMPI_COMM_WORLD_RANK"); // from OPENMPI
  if (sprocid) procid = atoi(sprocid); /* id0 is talking, id1++ is silent */
  slurmd_nodename=getenv("SLURMD_NODENAME");  /* nodename or "" */
  if (verbose>6) printf("DBG L%d host=%s\n",__LINE__,slurmd_nodename);
  if (!slurmd_nodename) // kautz18 geht erst nach export HOSTNAME, set, ~env
    slurmd_nodename=getenv("HOSTNAME");  /* nodename or "" CentOS5 */
  if (verbose>6) printf("DBG L%d host=%s\n",__LINE__,slurmd_nodename);
  // ToDo: /proc/sys/kernel/hostname = unistd.gethostname(char *,size)
  if (!slurmd_nodename)
    { static char s1[128]={0}; gethostname(s1,128); // 2018-03-18
      if (s1[0]) slurmd_nodename=s1; }
  if (!slurmd_nodename)
    slurmd_nodename=dummy_nodename; /* we dont want NULL-pointer "?" */
  if (verbose>6) printf("DBG L%d host=%s\n",__LINE__,slurmd_nodename);
  // 2018-03 atoi(sprocid) = SEGFAULT
  { char    *s1=getenv("SLURM_NPROCS");           /* slurm? */
    if (!s1) s1=getenv("OMPI_COMM_WORLD_SIZE");   /* openmpi? */
    num_nodes=(s1)?atoi(s1):1; }
  if (num_nodes==0) num_nodes=1;
  if (verbose>6) printf("DBG L%d\n",__LINE__);
  if (!procid) {
//SLURM_NNODES=1
//SLURM_NPROCS=1
//SLURM_NODELIST=scx-m0n0
    printf("mem_badram bad RAM detection 2012-2020.06 by JS %luMB l%d v%d\n",
      mb, n_loops, verbose);
    printf(" NUM_PROCS= %d\n", num_nodes);
  }
  if (procid<3) {
    printf(" SLURM_PROCID=%s = %d %s (skip3++)\n",
       sprocid, procid, slurmd_nodename); /* 0..n or "" */
    fflush(stdout);
  }
#if defined(__GNUC__) // 2020-06 for better documentation
    printf("Compiler: GNUC %d.%d.%d\n",
                  __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__);
#endif
  if (verbose>6) printf("DBG L%d\n",__LINE__);
#ifdef _OPENMP
#pragma omp parallel default(shared)
  omp_n= omp_get_num_threads();  /* must be called within parallel pragma */
  if(procid==0)printf(" using OpenMP OMP_NUM_THREADS= %d\n", omp_n);
  if(procid==1)printf(" using OpenMP OMP_NUM_THREADS= %d (slurm=1)\n", omp_n);
  // export OMP_NUM_THREADS=16;GOMP_CPU_AFFINITY="0-31" ./mem_badram 120000
  if(procid==0)printf(" using OpenMP GOMP_CPU_AFFINITY=\"%s\"\n",
    getenv("GOMP_CPU_AFFINITY"));
#endif
  chk_edac(procid,0,0,bad,&bad_nb); // print initial edac values, panic_on_ue shoud be 0
  if (!procid) {
    printf(" malloc %lu MiB (Mi=2^20) long= %d B = %d bit (arg1)\n",
           mb /* sumMB */, sl /* sizeofLong */, sl*8 /* bits */);
    // split into 1024 blocks of size mb*1024
    printf(" malloc 1024 blocks to %lu kiB = %lu * long (ki=2^10)\n", mb, num);
    printf(" starting, only every 128th block shown ...\n"); fflush(stdout);
  }
  fflush(stdout);
  for (i=0;i<1024;i++) { // 1024 memory blocks * 128MiB = 128GB
    bad[i] = 0; /* marked as clean */
    pa[i] = (ULong *) malloc(mb*1024l); /* alloc mb*kB (memoryMB/1024) */
    // posix_memalign((void **)pa[i], 64, mb*1024l); 
    // usleep(100000);
    if (procid==0 && ((i&127)==0 || !pa[i]))
      printf(" a=%016lx = %3luGB blk %4lu sz=%4lukB (alloc+ini)\n", // old: touch
             (ULong)pa[i], ((ULong)pa[i])>>30, i, mb);
    if (!pa[i]) { break; }  // malloc failed
    // touching j+=512 (*8B=4kB) is faster, but gives bad binding on openmp
    // cat /sys/devices/system/node/node[01]/numastat t100.TODO numa_hit
    //  see t100.node064
  }
 
  if ( mlockall(MCL_CURRENT|MCL_FUTURE) == -1 )
    printf("ERROR: mlockall failed (non-root?)\n"); /* root only */ 
#pragma omp parallel for default(shared) private(i,j) schedule(static)
  for (i=0;i<1024;i++) {
    //  TODO better parallelize outer loop over i=0..1023
//#pragma omp parallel for default(shared) private(j) schedule(static)
    for (j=0; j<num; j++) pa[i][j]=0; // single task faster if j+=512 (8B)
  } nb=i; // num blocks (serial)
//  if ( mlockall(MCL_CURRENT|MCL_FUTURE) == -1 )
//    printf("ERROR: mlockall failed (non-root?)\n"); /* root only */ 
  if (nb!=1024) 
    printf(" WARNING: only %lu blocks = %lu kiB allocated on %s\n",
           nb+1, (nb+1)*mb, slurmd_nodename);
  if (!nb) exit(1);
  { int ce0=chk_edac(procid,-1,0,bad,&bad_nb);if (ce0) edac_errors+=ce0; }
  for (ii=0;ii<n_loops;ii++) { // dflt 100 rounds
    int rc=0; 
    rnd1=rand(); // random fill pattern, ToDo better as seed for xorshift-fill
    rnd2=rand();
    rnd=(rnd1<<32) | rnd2;  // 64bit random
    if (/*ii!=0 &&*/ (ii&7)==0) rnd=0; /* 1803: ii!=0 */
    // printf(" rnd %016lx %016lx %016lx\n", rnd1, rnd2, rnd);
    t1=fstime(); // 16*(wr+rd)
    // return number bad blocks, wr+rd
    bad_nb+=(rc=chkmem(pa,nb,num,ii,rnd,procid,bad,bad_nb)); // 0=OK 16*invert+write+read+cmp
    t2=fstime();
    /* 1803.m34n12 breaks at invers loop, lot of false bad output */
    // if (rc && pa[0][0]!=rnd && pa[0][0]==~rnd) rnd=~rnd; /* after break even */
    /* 1804 log2 output */
    if (procid==0 && t1!=t2 && ((ii&(ii-1))==0 || ii<8)) /* avoid div0 */
      printf(" loop=%3d speed.wr: %10.3f GiB/s (%.0f MiB)\n",
             (int)ii, ((edac_blk_bitmap)?0.5:1.0)*
               INVLOOPs*1.*nb/1024./1024.*mb/(t2-t1), mb*1.*nb/1024.);
    // if (bad_nb>3 && bad_nb!=1024/*bug?*/) break; // 2018-03 ~brk, see c029
    // 2018-03 do not break loop, just skip bad blocks on further search
    usleep(60000); // 60ms (extra errors?)
    // getconf -a | grep -i cache # L1+L2_CACHE_SIZE+ASSOC=8+LINESIZE=64
    t1=fstime();
    bad_nb+= mem_rd(pa, nb, num, rnd, procid, bad, 32);
    t2=fstime(); // double
    if (procid==0 && t2-t1>0 && ((ii&(ii-1))==0 || ii<8)) // ii=0,1,2,4
      printf(" loop=%3d speed.r:  %10.3f GiB/s (%.0f MiB)\n",
       (int)ii, ((edac_blk_bitmap)?0.5:1.0)*
         1.*nb/1024./1024.*mb/(t2-t1), mb*1.*nb/1024.);
    // t100.n146 86-93GB/s=89.5+-4%, why not const?  2019-01
    // ToDo: within 65ms DRAM refresh do 130e3 uncached reads
    
    // latency: L1=1ns=light/30.5cm L2=3ns L3=12ns lMem=60ns rMem=100ns
    // ToDo: core1: full-L1-loop core2: access remote RAM or vv 
    t1=fstime();
      mem_rd_hammer(pa, nb, num, rnd, procid, bad, 32);
    t2=fstime();
    if (procid==0 && t2-t1>0 && ((ii&(ii-1))==0 || ii<8))
      printf(" loop=%3d speed.rh: %10.0f MB/s (%.3f KB)\n",
       (int)ii, 2048.*mb/1024./(t2-t1), mb*1.);
    bad_nb+= mem_rd(pa, nb, num, rnd, procid, bad, 32);
 
    // Todo 2nd read after some seconds (+partly write = rowhammer?)
  } // loop100
  for (bad_nb=i=0;i<nb;i++) // 2018-04-05 add set NULL on free, avoid segflt
    if (pa[i] && !bad[i]) { free((void *)pa[i]); pa[i]=0; } // free good blocks
         else if (bad[i]) bad_nb++; // count bad blocks
  if (bad_nb && bad_nb!=1024) {
    printf("%d bad blocks detected at task%d=%s, sleep + fgets ...\n",
      bad_nb, procid, slurmd_nodename);
    for (;;) { char s1[8]/*,rc=0*/; /* wait for input within gnu-screen */
      s1[0]='\0';if (!fgets(s1,8,stdin)) printf("ERROR fgets\n");
      printf("DBG: read char %04d %c\n", s1[0],
          ((s1[0]>='0' && s1[0]<='z')?s1[0]:'.'));
      if (s1[0]!='\0') { // simply just enter = "\n\0"
        for (bad_nb=i=0;i<nb;i++) if (pa[i]) bad[i]=0; /* reset counter */
        edac_blk_bitmap=0;
        for (ii=0; ii<n_loops;ii++) 
        /*rc=*/chkmem(pa,nb,num,ii,rnd,procid,bad,bad_nb);
        /* segfault? */
      } else sleep(10); // if(system("sleep 600"))break; // sleep forever
      // check every minute for read test of bad blk!? retrigger error
     // 2018-04-04 node168 2cpus 16*16GB-ECC-DIMMs ca. 30 ce-errors/13 loops
     //   block number is changing from run to run, constant within run
    }
  } // if bad_nb
  if (bad_nb==0 && edac_errors==0 && procid==0)
    printf("-- no error detected on procid=0 %s --\n", slurmd_nodename);
  if (edac_errors) printf("ERR edac_errors=%d at node=%s\n",
    edac_errors, slurmd_nodename);
  system("date +%F\\ %T");
  // ToDo: how to integrate that into the jobsystem? extra-shared-queue?
  return 0;
}
